import os
import importlib
from pathlib import Path
from typing import Optional, Union

from itkwasm import Image, PixelTypes, IntTypes, FloatTypes, BinaryFile

from itkwasm.pyodide import (
    to_js,
    to_py,
    js_resources
)

from .js_package import js_package

from .extension_to_image_io import extension_to_image_io
from .image_io_index import image_io_index

async def write_image_async(
    image: Image,
    serialized_image: os.PathLike,
    information_only: bool = False,
    use_compression: bool = False,
) -> None:
    """Write an itk-wasm Image to an image file format.

    :param image: Input image
    :type  image: Image

    :param serialized_image: Output image serialized in the file format.
    :type  serialized_image: str

    :param information_only: Only write image metadata -- do not write pixel data.
    :type  information_only: bool

    :param use_compression: Use compression in the written file
    :type  use_compression: bool

    :param serialized_image: Input image serialized in the file format
    :type  serialized_image: os.PathLike
    """
    js_module = await js_package.js_module
    web_worker = js_resources.web_worker

    kwargs = {}
    if information_only:
        kwargs["informationOnly"] = to_js(information_only)
    if use_compression:
        kwargs["useCompression"] = to_js(use_compression)

    extension = ''.join(Path(serialized_image).suffixes)

    io = None
    if extension in extension_to_image_io:
        func = f"{extension_to_image_io[extension]}WriteImage"
        io = getattr(js_module, func)
    else:
        for ioname in image_io_index:
            func = f"{ioname}WriteImage"
            io = getattr(js_module, func)
            outputs = await io(to_js(image), to_js(serialized_image), webWorker=web_worker, noCopy=True, **kwargs)
            outputs_object_map = outputs.as_object_map()
            web_worker = outputs_object_map['webWorker']
            js_resources.web_worker = web_worker
            could_write = to_py(outputs_object_map['couldWrite'])
            if could_write:
                to_py(outputs_object_map['serializedImage'])
                return

    if io is None:
        raise RuntimeError(f"Could not find an image writer for {extension}")

    outputs = await io(to_js(image), to_js(serialized_image), webWorker=web_worker, noCopy=True, **kwargs)
    outputs_object_map = outputs.as_object_map()
    web_worker = outputs_object_map['webWorker']
    js_resources.web_worker = web_worker
    could_write = to_py(outputs_object_map['couldWrite'])

    if not could_write:
        raise RuntimeError(f"Could not write {serialized_image}")

    to_py(outputs_object_map['serializedImage'])

async def imwrite_async(
    image: Image,
    serialized_image: os.PathLike,
    information_only: bool = False,
    use_compression: bool = False,
) -> None:
    return write_image_async(image, serialized_image, information_only=information_only, use_compression=use_compression)

imwrite_async.__doc__ = f"""{write_image_async.__doc__}
    Alias for write_image.
    """



import pytest
import sys
import glob

if sys.version_info < (3,10):
    pytest.skip("Skipping pyodide tests on older Python", allow_module_level=True)

from pytest_pyodide import run_in_pyodide

from itkwasm_image_io_emscripten import __version__ as test_package_version

@pytest.fixture
def package_wheel():
    return f"itkwasm_image_io_emscripten-{test_package_version}-py3-none-any.whl"

@pytest.fixture
def input_data():
    from pathlib import Path
    input_base_path = Path(__file__).parent.parent / 'test' / 'data'
    test_files = list(input_base_path.glob('*'))
    data = {}
    for test_file in test_files:
        with open(test_file, 'rb') as f:
            data[test_file.name] = f.read()
    return data



image_io_index = [
  'png',
  'meta',
  'tiff',
  'nifti',
  'jpeg',
  'nrrd',
  'vtk',
  'bmp',
  'hdf5',
  'minc',
  'mrc',
  'lsm',
  'mgh',
  'bio_rad',
  'gipl',
  'ge_adw',
  'ge4',
  'ge5',
  'gdcm',
  'scanco',
  'wasm',
  'wasm_zstd',
]




import os
from typing import Optional, Union
from pathlib import Path

from itkwasm import (
    Image,
    PixelTypes,
    IntTypes,
    FloatTypes,
    BinaryFile,
    cast_image,
)

from .js_package import js_package

from itkwasm.pyodide import (
    to_js,
    to_py,
    js_resources
)

from .extension_to_image_io import extension_to_image_io
from .image_io_index import image_io_index

async def read_image_async(
    serialized_image: os.PathLike,
    information_only: bool = False,
    pixel_type: Optional[PixelTypes]=None,
    component_type: Optional[Union[IntTypes, FloatTypes]]=None,
) -> Image:
    """Read an image file format and convert it to the itk-wasm file format

    :param serialized_image: Input image serialized in the file format
    :type  serialized_image: os.PathLike

    :param information_only: Only read image metadata -- do not read pixel data.
    :type  information_only: bool

    :param pixel_type: Pixel type to cast to.
    :type  pixel_Type: Optional[PixelTypes]

    :param component_type: Component type to cast to.
    :type  component_type: Optional[Union[IntTypes, FloatTypes]]

    :return: Output image
    :rtype:  Image
    """
    js_module = await js_package.js_module
    web_worker = js_resources.web_worker

    kwargs = {}
    if information_only:
        kwargs["informationOnly"] = to_js(information_only)

    extension = ''.join(Path(serialized_image).suffixes)

    io = None
    if extension in extension_to_image_io:
        func = f"{extension_to_image_io[extension]}ReadImage"
        io = getattr(js_module, func)
    else:
        for ioname in image_io_index:
            func = f"{ioname}ReadImage"
            io = getattr(js_module, func)
            outputs = await io(to_js(BinaryFile(serialized_image)), webWorker=web_worker, noCopy=True, **kwargs)
            outputs_object_map = outputs.as_object_map()
            web_worker = outputs_object_map['webWorker']
            js_resources.web_worker = web_worker
            could_read = to_py(outputs_object_map['couldRead'])
            if could_read:
                image = to_py(outputs_object_map['image'])
                if pixel_type or component_type:
                    image = cast_image(image, pixel_type=pixel_type, component_type=component_type)
                return image

    if io is None:
        raise RuntimeError(f"Could not find an image reader for {extension}")

    outputs = await io(to_js(BinaryFile(serialized_image)), webWorker=web_worker, noCopy=True, **kwargs)
    outputs_object_map = outputs.as_object_map()
    web_worker = outputs_object_map['webWorker']
    could_read = to_py(outputs_object_map['couldRead'])

    if not could_read:
        raise RuntimeError(f"Could not read {serialized_image}")

    js_resources.web_worker = web_worker

    image = to_py(outputs_object_map['image'])

    if pixel_type or component_type:
        image = cast_image(image, pixel_type=pixel_type, component_type=component_type)
    return image

async def imread_async(
    serialized_image: os.PathLike,
    information_only: bool = False,
    pixel_type: Optional[PixelTypes]=None,
    component_type: Optional[Union[IntTypes, FloatTypes]]=None,
) -> Image:
    return await read_image_async(serialized_image, information_only=information_only, pixel_type=pixel_type, component_type=component_type)

imread_async.__doc__ = f"""{read_image_async.__doc__}
    Alias for read_image_async.
    """



from collections import OrderedDict

extension_to_image_io = OrderedDict([
  ('.bmp', 'bmp'),

  ('.dcm', 'gdcm'),

  ('.gipl', 'gipl'),
  ('.gipl.gz', 'gipl'),

  ('.hdf5', 'hdf5'),

  ('.jpg', 'jpeg'),
  ('.jpeg', 'jpeg'),

  ('.iwi', 'wasm'),
  ('.iwi.cbor', 'wasm'),
  ('.iwi.cbor.zst', 'wasmZstd'),

  ('.lsm', 'lsm'),

  ('.mnc', 'mnc'),
  ('.mnc.gz', 'mnc'),
  ('.mnc2', 'mnc'),

  ('.mgh', 'mgh'),
  ('.mgz', 'mgh'),
  ('.mgh.gz', 'mgh'),

  ('.mha', 'meta'),
  ('.mhd', 'meta'),

  ('.mrc', 'mrc'),

  ('.nia', 'nifti'),
  ('.nii', 'nifti'),
  ('.nii.gz', 'nifti'),
  ('.hdr', 'nifti'),

  ('.nrrd', 'nrrd'),
  ('.nhdr', 'nrrd'),

  ('.png', 'png'),

  ('.pic', 'bioRad'),

  ('.tif', 'tiff'),
  ('.tiff', 'tiff'),

  ('.vtk', 'vtk'),

  ('.isq', 'scanco'),
  ('.aim', 'scanco'),

  ('.fdf', 'fdf'),
])



import pytest
import sys

if sys.version_info < (3,10):
    pytest.skip("Skipping pyodide tests on older Python", allow_module_level=True)

from pytest_pyodide import run_in_pyodide

from itkwasm_image_io_emscripten import __version__ as test_package_version

@pytest.fixture
def package_wheel():
    return f"itkwasm_image_io_emscripten-{test_package_version}-py3-none-any.whl"

@pytest.fixture
def input_data():
    from pathlib import Path
    input_base_path = Path('..', '..', 'test', 'data')
    test_files = [
        Path('input') / 'cthead1.png',
        Path('input') / 'biorad.pic',
        Path('input') / 'brainweb165a10f17.mha',
    ]
    data = {}
    for f in test_files:
        with open(input_base_path / f, 'rb') as fp:
            data[str(f.name)] = fp.read()
    return data


import sys

if sys.version_info < (3,10):
    pytest.skip("Skipping pyodide tests on older Python", allow_module_level=True)

from pytest_pyodide import run_in_pyodide
from .fixtures import package_wheel, input_data

@run_in_pyodide(packages=['micropip'])
async def test_metaimage_async(selenium, package_wheel, input_data):
    import micropip
    await micropip.install(package_wheel)
    def write_input_data_to_fs(input_data, filename):
        with open(filename, 'wb') as fp:
            fp.write(input_data[filename])

    from pathlib import Path

    from itkwasm_image_io_emscripten import meta_read_image_async, meta_write_image_async

    test_input_file_path = 'brainweb165a10f17.mha'
    test_output_file_path = 'meta-image-test-brainweb165a10f17.mha'

    def verify_image(image):
        assert image.imageType.dimension == 3
        assert image.imageType.componentType == "uint8"
        assert image.imageType.pixelType == "Scalar"
        assert image.imageType.components == 1
        assert image.origin[0] == 0.0
        assert image.origin[1] == 0.0
        assert image.origin[2] == 0.0
        assert image.spacing[0] == 1.0
        assert image.spacing[1] == 1.0
        assert image.spacing[2] == 1.0
        assert image.direction[0, 0] == 1.0
        assert image.direction[0, 1] == 0.0
        assert image.direction[0, 2] == 0.0
        assert image.direction[1, 0] == 0.0
        assert image.direction[1, 1] == 1.0
        assert image.direction[1, 2] == 0.0
        assert image.direction[2, 0] == 0.0
        assert image.direction[2, 1] == 0.0
        assert image.direction[2, 2] == 1.0
        assert image.size[0] == 181
        assert image.size[1] == 217
        assert image.size[2] == 180
        assert image.data.ravel()[0] == 5
        assert image.data.ravel()[1] == 8
        assert image.data.ravel()[2] == 2

    write_input_data_to_fs(input_data, test_input_file_path)

    assert Path(test_input_file_path).exists()

    could_read, image = await meta_read_image_async(test_input_file_path)
    assert could_read
    verify_image(image)

    use_compression = False
    could_write = await meta_write_image_async(image, test_output_file_path, use_compression)
    assert could_write

    could_read, image = await meta_read_image_async(test_output_file_path)
    assert could_read
    verify_image(image)


# Generated file. To retain edits, remove this comment.

from pathlib import Path
import os
from typing import Dict, Tuple, Optional, List, Any

from .js_package import js_package

from itkwasm.pyodide import (
    to_js,
    to_py,
    js_resources
)
from itkwasm import (
    InterfaceTypes,
    Image,
    BinaryFile,
)

async def scanco_write_image_async(
    image: Image,
    serialized_image: str,
    information_only: bool = False,
    use_compression: bool = False,
) -> Tuple[Any]:
    """Write an itk-wasm file format converted to an image file format

    :param image: Input image
    :type  image: Image

    :param serialized_image: Output image serialized in the file format.
    :type  serialized_image: str

    :param information_only: Only write image metadata -- do not write pixel data.
    :type  information_only: bool

    :param use_compression: Use compression in the written file
    :type  use_compression: bool

    :return: Whether the input could be written. If false, the output image is not valid.
    :rtype:  Any
    """
    js_module = await js_package.js_module
    web_worker = js_resources.web_worker

    kwargs = {}
    if information_only:
        kwargs["informationOnly"] = to_js(information_only)
    if use_compression:
        kwargs["useCompression"] = to_js(use_compression)

    outputs = await js_module.scancoWriteImage(to_js(image), to_js(serialized_image), webWorker=web_worker, noCopy=True, **kwargs)

    output_web_worker = None
    output_list = []
    outputs_object_map = outputs.as_object_map()
    for output_name in outputs.object_keys():
        if output_name == 'webWorker':
            output_web_worker = outputs_object_map[output_name]
        else:
            output_list.append(to_py(outputs_object_map[output_name]))

    js_resources.web_worker = output_web_worker

    if len(output_list) == 1:
        return output_list[0]
    return tuple(output_list)



import sys

if sys.version_info < (3,10):
    pytest.skip("Skipping pyodide tests on older Python", allow_module_level=True)

import pytest
from pytest_pyodide import run_in_pyodide
from .fixtures import package_wheel, input_data

@pytest.mark.driver_timeout(30)
@run_in_pyodide(packages=['micropip'])
async def test_bio_rad_async(selenium, package_wheel, input_data):
    import micropip
    await micropip.install(package_wheel)
    def write_input_data_to_fs(input_data, filename):
        with open(filename, 'wb') as fp:
            fp.write(input_data[filename])

    from pathlib import Path

    from itkwasm_image_io_emscripten import bio_rad_read_image_async, bio_rad_write_image_async

    def verify_image(image):
        assert image.imageType.dimension == 2
        assert image.imageType.componentType == "uint8"
        assert image.imageType.pixelType == "Scalar"
        assert image.imageType.components == 1
        assert image.origin[0] == 0.0
        assert image.origin[1] == 0.0
        assert image.spacing[0] == 0.06000000238418579
        assert image.spacing[1] == 0.06000000238418579
        assert image.direction[0, 0] == 1.0
        assert image.direction[0, 1] == 0.0
        assert image.direction[1, 0] == 0.0
        assert image.direction[1, 1] == 1.0
        assert image.size[0] == 768
        assert image.size[1] == 512
        assert image.data.shape[1] == 768
        assert image.data.shape[0] == 512
        assert image.data.ravel()[1000] == 27

    test_file_path = 'biorad.pic'
    write_input_data_to_fs(input_data, test_file_path)

    assert Path(test_file_path).exists()

    could_read, image = await bio_rad_read_image_async(test_file_path)
    assert could_read
    verify_image(image)

    test_output_file_path = 'biorad_out.pic'

    use_compression = False
    could_write = await bio_rad_write_image_async(image, test_output_file_path, use_compression)
    assert could_write

    could_read, image = await bio_rad_read_image_async(test_output_file_path)
    assert could_read
    verify_image(image)